# Claudio Perez
"""
- glTF uses a right-handed coordinate system, that is, the cross product of +X and +Y yields +Z.
- glTF defines +Y as up.
- The front of a glTF asset faces +Z.
- All angles are in radians.
- Positive rotation is counterclockwise.
- Rotations are given as quaternions stored as a tuple (x,y,z,w),
  where the w-component is the cosine of half of the rotation angle.
  For example, the quaternion [ 0.259, 0.0, 0.0, 0.966 ] describes a rotation
  about 30 degrees, around the x-axis.
"""
import sees
from pathlib import Path
import warnings
import numpy as np
import pygltflib
from .canvas import Canvas, NodeStyle, MeshStyle, LineStyle, DrawStyle

GLTF_T = {
    "float32": pygltflib.FLOAT,
    "uint8":   pygltflib.UNSIGNED_BYTE,
    "uint16":  pygltflib.UNSIGNED_SHORT,
}

def _split(a, value, source=None):
    # like str.split, but for arrays
    if source is None:
        source = a
    if np.isnan(value):
        idx = np.where(~np.isnan(source))[0]
    else:
        idx = np.where(source != value)[0]
    return np.split(a[idx],np.where(np.diff(idx)!=1)[0]+1)

class GltfLibCanvas(Canvas):
    vertical = 2

    def __init__(self, config=None):
        self.config = config

        #                          x, y, z, scalar
        self._rotation = [-0.7071068, 0, 0, 0.7071068]
        # equivalent rotation matrix:
        self._rotation_matrix = np.array([[1,  0, 0],
                                          [0,  0, 1],
                                          [0, -1, 0]])

        self.index_t = "uint16"
        self.float_t = "float32"

        self.gltf = pygltflib.GLTF2(
            scene=0,
            scenes=[pygltflib.Scene(nodes=[])],
            nodes=[],
            meshes=[],
            accessors=[],
            materials=[
                pygltflib.Material(
                    name="black",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[0, 0, 0, 1]
                    )
                ),
                pygltflib.Material(
                    name="white",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[1, 1, 1, 1]
                    )
                ),
                pygltflib.Material(
                    name="red",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[1, 0, 0, 1]
                    )
                ),
                pygltflib.Material(
                    name="green",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[0, 1, 0, 1]
                    )
                ),
                pygltflib.Material(
                    name="blue",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[0, 0, 1, 1]
                    )
                ),
                pygltflib.Material(
                    name="gray",
                    doubleSided=True,
                    alphaMode=pygltflib.MASK,
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        baseColorFactor=[0.9, 0.9, 0.9, 1]
                    )
                ),
                pygltflib.Material(
                    name="metal",
                    doubleSided=True,
#                   alphaMode=pygltflib.MASK,
                    occlusionTexture=pygltflib.OcclusionTextureInfo(index=1),
#                   emissiveFactor=[0.8,0.8,0.8],
                    pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                        metallicFactor=0.0,
                        roughnessFactor=1.0,
#                       baseColorFactor=[0.8, 0.8, 0.8, 1],
                        baseColorTexture=pygltflib.TextureInfo(index=0),
                        metallicRoughnessTexture=pygltflib.TextureInfo(index=1),
                    )
                ),
            ],
            bufferViews=[],
            buffers=[pygltflib.Buffer(byteLength=0)],
        )
        self.gltf._glb_data = bytes()


        # Map pairs of (color, alpha) to material's index in material list
        self._color = {(m.name,m.pbrMetallicRoughness.baseColorFactor[3]): i
                       for i,m in enumerate(self.gltf.materials)}

        #
        # load assets for steel material
        #
        try:
#           for i,file in enumerate(("baseColor.png", "occlusionRoughnessMetallic.png")):
            for i,file in enumerate(("rust.jpg", "occlusionRoughnessMetallic.png")):
                path  = str(sees.assets/"metal"/file)
                image = pygltflib.Image()
                image.uri = path
                self.gltf.images.append(image)
                self.gltf.textures.append(pygltflib.Texture(source=i, name=path))

            self.gltf.convert_images(pygltflib.ImageFormat.DATAURI)
        except:
            warnings.warn(f"Failed to load assets from {sees.assets}.")

    def _init_nodes(self, style: NodeStyle):
        #
        #
        #
        index_t = "uint8"
        points = style.scale*np.array(
            [
                [-1.0, -1.0,  1.0],
                [ 1.0, -1.0,  1.0],
                [-1.0,  1.0,  1.0],
                [ 1.0,  1.0,  1.0],
                [ 1.0, -1.0, -1.0],
                [-1.0, -1.0, -1.0],
                [ 1.0,  1.0, -1.0],
                [-1.0,  1.0, -1.0],
            ],
            dtype=self.float_t,
        )/10

        triangles = np.array(
            [
                [0, 1, 2],
                [3, 2, 1],
                [1, 0, 4],
                [5, 4, 0],
                [3, 1, 6],
                [4, 6, 1],
                [2, 3, 7],
                [6, 7, 3],
                [0, 2, 5],
                [7, 5, 2],
                [5, 7, 4],
                [6, 4, 7],
            ],
            dtype=index_t,
        )
        triangles_binary_blob = triangles.flatten().tobytes()
        points_binary_blob = points.tobytes()

        self.gltf.accessors.extend([
            pygltflib.Accessor(
                bufferView=self._push_data(triangles_binary_blob,
                                           pygltflib.ELEMENT_ARRAY_BUFFER),
                componentType=GLTF_T[index_t],
                count=triangles.size,
                type=pygltflib.SCALAR,
                max=[int(triangles.max())],
                min=[int(triangles.min())],
            ),
            pygltflib.Accessor(
                bufferView=self._push_data(points_binary_blob,
                                           pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T[self.float_t],
                count=len(points),
                type=pygltflib.VEC3,
                max=points.max(axis=0).tolist(),
                min=points.min(axis=0).tolist(),
            )
        ])


        indices_access = len(self.gltf.accessors)-2 # indices
        points_access  = len(self.gltf.accessors)-1 # points
        self.gltf.meshes.append(
               pygltflib.Mesh(
                 primitives=[
                     pygltflib.Primitive(
                         mode=pygltflib.TRIANGLES,
                         attributes=pygltflib.Attributes(POSITION=points_access),
                         material=self._get_material(style),
                         indices=indices_access
                     )
                 ]
               )
        )

        self._node_mesh = len(self.gltf.meshes) - 1

    def _use_asset(self, name, scale, rotation, material):
        pass

    def plot_nodes(self, vertices, label = None, style=None, data=None, **kwds):

        if not hasattr(self, "_node_mesh"):
            self._init_nodes(style or NodeStyle())


        for coord in vertices:
            self.gltf.nodes.append(pygltflib.Node(
                    mesh=self._node_mesh,
                    #rotation=self._rotation,
                    translation=(self._rotation_matrix@coord).tolist(),
                )
            )
            self.gltf.scenes[0].nodes.append(len(self.gltf.nodes)-1)



    def _get_material(self, style: DrawStyle)->int:
        if hasattr(style,"alpha"):
            alpha = style.alpha
        else:
            alpha = 1.0

        color = style.color

        if (color, alpha) in self._color:
            return self._color[(color,alpha)]

        elif isinstance(color, str) and color[0] == "#":
            # Remove leading hash in hex
            hx  = color.lstrip("#")
            # Convert hex to RGB
            rgb = [int(hx[i:i+2], 16)/255 for i in (0, 2, 4)]

        elif isinstance(color, tuple):
            rgb = color

        else:
            raise TypeError("Unexpected type for color")

        # Store index for new material
        self._color[(color, alpha)] = len(self.gltf.materials)
        # Create and add new material
        self.gltf.materials.append(
            pygltflib.Material(
                name=str(color),
                doubleSided=True,
                alphaMode=pygltflib.MASK,
                pbrMetallicRoughness=pygltflib.PbrMetallicRoughness(
                    baseColorFactor=[*rgb, alpha]
                )
            ),
        )
        return self._color[(color, alpha)]


    def _push_data(self, data, target, byteStride=None)->int:
        self.gltf.bufferViews.append(
                pygltflib.BufferView(
                    buffer=0,
                    byteStride=byteStride,
                    byteOffset=self.gltf.buffers[0].byteLength,
                    byteLength=len(data),
                    target=target,
                )
        )

        self.gltf._glb_data += data
        self.gltf.buffers[0].byteLength += len(data)
        return len(self.gltf.bufferViews)-1


    def plot_lines(self, vertices, indices=None, style: LineStyle=None, **kwds):
        material = self._get_material(style or LineStyle())


        # vertices is given with nans separating line groups, but
        # GLTF does not accept nans so we have to filter these
        # out, and add distinct meshes for each line group
        assert np.all(np.isnan(vertices[np.isnan(vertices[:,0]), :]))
        points  = np.array(vertices[~np.isnan(vertices[:,0]),:], dtype=self.float_t)
        points_buffer = self._push_data(points.tobytes(), pygltflib.ARRAY_BUFFER)

        self.gltf.accessors.append(
            pygltflib.Accessor(
                bufferView=points_buffer,
                componentType=GLTF_T[self.float_t],
                count=len(points),
                type=pygltflib.VEC3,
                max=points.max(axis=0).tolist(),
                min=points.min(axis=0).tolist(),
            )
        )
        points_access = len(self.gltf.accessors) - 1

        if indices is None:
            indices_ = _split(np.arange(len(vertices), dtype=self.index_t), np.nan, vertices[:,0])
        else:
            indices_ = indices

        print(vertices)
        print(indices_)

        for indices in indices_:
            # here, n adjusts indices by the number of nan rows that were removed so far
            n  = sum(np.isnan(vertices[:indices[0],0]))
            indices_array = indices - n
            indices_binary_blob = indices_array.tobytes()

            if len(indices_array) <= 1:
                print(indices_array)
                continue

            self.gltf.accessors.extend([
                pygltflib.Accessor(
                    bufferView=self._push_data(indices_binary_blob,
                                               pygltflib.ELEMENT_ARRAY_BUFFER),
                    componentType=GLTF_T[self.index_t],
                    count=indices_array.size,
                    type=pygltflib.SCALAR,
                    max=[int(indices_array.max())],
                    min=[int(indices_array.min())],
                )
            ])
            self.gltf.meshes.append(
                   pygltflib.Mesh(
                     primitives=[
                         pygltflib.Primitive(
                             mode=pygltflib.LINES,
                             attributes=pygltflib.Attributes(POSITION=points_access),
                             material=material,
                             # most recently added accessor
                             indices=len(self.gltf.accessors)-1,
                             # TODO: use this mechanism to add annotation data
                             extras={},
                         )
                     ]
                   )
            )

            self.gltf.nodes.append(pygltflib.Node(
                    mesh=len(self.gltf.meshes)-1,
                    rotation=self._rotation
                )
            )
            self.gltf.scenes[0].nodes.append(len(self.gltf.nodes)-1)


    def plot_mesh(self, vertices, triangles, local_coords=None, style=None, **kwds):

        material = self._get_material(style or MeshStyle())
        points    = np.array(vertices, dtype=self.float_t)
        triangles = np.array(triangles,dtype=self.index_t)

        triangles_binary_blob = triangles.flatten().tobytes()
        points_binary_blob = points.tobytes()

        self.gltf.accessors.extend([
            pygltflib.Accessor(
                bufferView=self._push_data(triangles_binary_blob, pygltflib.ELEMENT_ARRAY_BUFFER),
                componentType=GLTF_T[self.index_t],
                count=triangles.size,
                type=pygltflib.SCALAR,
                max=[int(triangles.max())],
                min=[int(triangles.min())],
            ),
            pygltflib.Accessor(
                bufferView=self._push_data(points_binary_blob, pygltflib.ARRAY_BUFFER),
                componentType=GLTF_T[self.float_t],
                count=len(points),
                type=pygltflib.VEC3,
                max=points.max(axis=0).tolist(),
                min=points.min(axis=0).tolist(),
            )
        ])

        index_access = len(self.gltf.accessors)-2
        point_access = len(self.gltf.accessors)-1

        self.gltf.meshes.append(
               pygltflib.Mesh(
                 primitives=[
                     pygltflib.Primitive(
                         mode=pygltflib.TRIANGLES,
                         attributes=pygltflib.Attributes(POSITION=point_access),
                         material=material,
                         indices=index_access
                     )
                 ]
               )
        )
        if local_coords is not None:
            locoor = np.array(local_coords, dtype=self.float_t)
            locoor_binary_blob = locoor.tobytes()
            self.gltf.accessors.extend([
                pygltflib.Accessor(
                    bufferView=self._push_data(locoor_binary_blob, pygltflib.ARRAY_BUFFER),
                    componentType=GLTF_T[self.float_t],
                    count=len(locoor),
                    type=pygltflib.VEC2,
                    max=locoor.max(axis=0).tolist(),
                    min=locoor.min(axis=0).tolist(),
                )
            ])
            self.gltf.meshes[-1].primitives[0].attributes.TEXCOORD_0 = len(self.gltf.accessors) -1

        self.gltf.nodes.append(pygltflib.Node(
                mesh=len(self.gltf.meshes)-1,
                rotation=self._rotation
            )
        )
        self.gltf.scenes[0].nodes.append(len(self.gltf.nodes)-1)


    def to_glb(self)->bytes:
        return b"".join(self.gltf.save_to_bytes())

    def write(self, filename=None):

        self.gltf.save(filename)

#       if "glb" in filename[-3:]:
#           glb = b"".join(self.gltf.save_to_bytes())
#           with open(filename,"wb+") as f:
#               f.write(glb)




"""
# Synopsis

>`render.py [<options>] <model-file>`

>**Chrystal Chern**, and **Claudio Perez**


This script plots the geometry of a structural
model given a SAM JSON file.


## Matlab
In order to install the Matlab bindings, open Matlab in a
directory containing the files `render.py` and `render.m`,
and run the following command in the Matlab interpreter:

    render --install

Once this process is complete, the command `render` can be
called from Matlab, just as described below for the command
line.

# Usage
This script can be used either as a module, or as a command
line utility. When invoked from the command line on
**Windows**, {NAME} should be `python -m render`. For example:

    python -m render model.json --axes 2 --view elev

"""
import sys

#import sees
from sees import RenderError, render
from sees.cli import parse_args


NAME="sees"

def main(argv):
    config = parse_args(argv)

    if config is None:
        # An error occured in parsing the command line; 
        # assume a message was printed
        sys.exit()

    try:
        artist = render(**config)

        # write plot to file if output file name provided
        if config["write_file"]:
            artist.save(config["write_file"])
            return

        # Otherwise either create popup, or start server
        elif hasattr(artist.canvas, "popup"):
            artist.canvas.popup()

        elif hasattr(artist.canvas, "to_glb"):
            import sees.server
            viewer = config["viewer_config"].get("name", None)
            port = config["server_config"].get("port", None)
            server = sees.server.Server(glb=artist.canvas.to_glb(),
                                        viewer=viewer)
            server.run(port=port)

        elif hasattr(artist.canvas, "to_html"):
            import sees.server
            port = config["server_config"].get("port", None)
            server = sees.server.Server(html=artist.canvas.to_html())
            server.run(port=port)

    except (FileNotFoundError, RenderError) as e:
        # Catch expected errors to avoid printing an ugly/unnecessary stack trace.
        print(e, file=sys.stderr)
        print("         Run '{NAME} --help' for more information".format(NAME=NAME), file=sys.stderr)
        sys.exit()


if __name__ == "__main__":
    main(sys.argv)




import os
import sys
from pathlib import Path

from .config import Config, apply_config
from .frame import FrameArtist

class RenderError(Exception): pass

assets = Path(__file__).parents[0]/"assets/"

def Canvas(subplots=None, backend=None):
    pass


def serve(artist, viewer="mv", port=None):
    import sees.server
    if hasattr(artist.canvas, "to_glb"):
        server = sees.server.Server(glb=artist.canvas.to_glb(),
                                    viewer=viewer)
        server.run(port=port)

    elif hasattr(artist.canvas, "to_html"):
        server = sees.server.Server(html=artist.canvas.to_html())
        server.run(port=port)

def render(sam_file, res_file=None, noshow=False, ndf=6,
           artist = None, #: str|"Artist" = None,
           **opts):

    import sees.model

    # Configuration is determined by successively layering
    # from sources with the following priorities:
    #      defaults < file configs < kwds 

    config = Config()

    if sam_file is None:
        raise RenderError("ERROR -- expected required argument <sam-file>")

    # Read model data
    if isinstance(sam_file, (str, Path)):
        model_data = sees.model.read_model(sam_file)

    elif hasattr(sam_file, "asdict"):
        # Assuming an opensees.openseespy.Model
        model_data = sam_file.asdict()

    elif hasattr(sam_file, "read"):
        model_data = sees.model.read_model(sam_file)

    elif isinstance(sam_file, tuple):
        # TODO: (nodes, cells)
        pass

    elif not isinstance(sam_file, dict):
        model_data = sees.model.read_model(sam_file)

    else:
        model_data = sam_file

    if "RendererConfiguration" in model_data:
        apply_config(model_data["RendererConfiguration"], config)

    apply_config(opts, config)

    #
    # Create Artist
    #
    # A Model is created from model_data by the artist
    # so that the artist can inform it how to transform
    # things if neccessary.
    if artist is None:
        artist = "frame"

    if artist == "frame":
        artist = FrameArtist(model_data, ndf=ndf, **config)


    #
    # Read and process displacements 
    #
    if res_file is not None:
        artist.add_state(res_file,
                         scale=config["scale"],
                         only=config["mode_num"])

    elif config["displ"] is not None:
        pass
        # TODO: reimplement point displacements
        # cases = [artist.add_point_displacements(config["displ"], scale=config["scale"])]

    if "Displacements" in model_data:
        cases.extend(artist.add_state(model_data["Displacements"],
                                        scale=config["scale"],
                                        only=config["mode_num"]))

    artist.draw()

    return artist




# Claudio Perez
# Summer 2024
import textwrap
from pathlib import Path

class Viewer:
    def __init__(self, viewer=None, src=None):
        self._glbsrc = "./model.glb" if src is None else src
        self._viewer = viewer if viewer is not None else "mv"

    def get_html(self):
        if self._viewer == "three-160":
            with open(Path(__file__).parents[0]/"gltf.html", "r") as f:
                return f.read()

        elif self._viewer == "three-130":
            with open(Path(__file__).parents[0]/"index.html", "r") as f:
                return f.read()

        elif self._viewer == "mv":
            return textwrap.dedent(f"""
          <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN"
            "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
          <html xmlns="http://www.w3.org/1999/xhtml" lang="en">
          <head>
            <meta charset="utf-8">
            <script type="module" src="https://ajax.googleapis.com/ajax/libs/model-viewer/3.5.0/model-viewer.min.js">
            </script>
          </head>
          <body>
            <model-viewer alt="rendering"
                          src="{self._glbsrc}"
                          ar
                          style="width: 100%; height: 1000px;"
                          shadow-intensity="1" camera-controls touch-action="pan-y">
            </model-viewer>
          </body>
          </html>
        """)







# render(model, [state], [canvas=plotly|matplotlib|gltf])
# 
#     Artist(model, canvas)
#         model = Model(model)
# 
#         for elem in model:
#             ... stich
#             canvas.add_mesh(mesh)















#
# sees
#

def render(model: dict | meshio.Mesh | ,
           state, canvas=None, artist=None) -> Artist: ...


def read_model():
    pass

def create_model():
    pass

class Artist: # frame.FrameArtist | plane.PlaneArtist | patch.PatchArtist
    def __init__(self, model, state=None, canvas=None): ...

#
# sees.canvas
#

def figure(splts, type=None) -> Canvas: ...

class Canvas: # canvas.Matplotlib
    def add_curves(self): ...
    def add_origin(self): ...

    def add_planes(self): ...



class Model:
    def __init__(self, data, model): ...


class State:
    def __init__(self, data, model): ...

    def node_array(self): ...

    def cell_array(self): ...



class Series: pass




# Configuring
#============
# The following configuration options are available:
from collections import defaultdict

Config = lambda : {
  "model_config":  {},
  "canvas_config": {}, # kwds for canvas
  "viewer_config": {},
  "server_config": {},

  "show_objects":  ["frames.displ", "nodes", "legend", "elastica", "reference"],
  "mode_num"    :  None,
  "hide_objects":  ["origin"],
  "sam_file":      None,
  "res_file":      None,
  "write_file":    None,
  "displ":         defaultdict(list),
  "scale":         1.0,
  "vert":          2,
  "view":          "iso",

  "camera": {
      "view": "iso",               # iso | plan| elev[ation] | sect[ion]
      "projection": "orthographic" # perspective | orthographic
  },

  # Artist
  "displacements": {"scale": 1.0, "color": "#660505"},

  "objects": {
      "origin": {"color": "black", "scale": 1.0},
      "frames" : {
          "displaced": {"color": "red", "npoints": 20}
      },
      "nodes": {
          "scale": 1.0,
          "default": {"size": 3, "color": "#000000"},
          "displaced" : {},
          "fixed"  : {},
      },
      "sections": {"scale": 1.0}
  },

  # Canvas
  "canvas":        "matplotlib",
  "save_options": {
      # Options for when writing to an HTML file.
      "html": {
          "include_plotlyjs": True,
          "include_mathjax" : "cdn",
          "full_html"       : True
      }
  }
}

def apply_config(conf, opts):
    for k,v in conf.items():
        if isinstance(v,dict):
            apply_config(v, opts[k])
        else:
            opts[k] = v




import sys
from collections import defaultdict


import numpy as np
Array = np.ndarray
from scipy.linalg import block_diag

from sees.model  import FrameModel
from sees.state  import read_state, State, BasicState
from sees.config import Config, apply_config
from sees.canvas.canvas import LineStyle, NodeStyle


# Data shaping / Misc.
#----------------------------------------------------

# The following functions are used for reshaping data
# and carrying out other miscellaneous operations.

def _is_frame(el):
    return     "beam" in el["type"].lower() \
            or "dfrm" in el["type"].lower()


# Kinematics
#----------------------------------------------------

# The following functions implement various kinematic
# relations for standard frame models.

def elastic_curve(x: Array, v: list, L:float)->Array:
    "compute points along Euler's elastica"
    if len(v) == 2:
        ui, uj, (vi, vj) = 0.0, 0.0, v
    else:
        ui, vi, uj, vj = v
    xi = x/L                        # local coordinate
    N1 = 1.-3.*xi**2+2.*xi**3
    N2 = L*(xi-2.*xi**2+xi**3)
    N3 = 3.*xi**2-2*xi**3
    N4 = L*(xi**3-xi**2)
    y = ui*N1 + vi*N2 + uj*N3 + vj*N4
    return y.flatten()

def elastic_tangent(x: Array, v: list, L: float)->Array:
    if len(v) == 2:
        ui, uj, (vi, vj) = 0.0, 0.0, v
    else:
        ui, vi, uj, vj = v
    xi = x/L
    M3 = 1 - xi
    M4 = 6/L*(xi-xi**2)
    M5 = 1 - 4*xi+3*xi**2
    M6 = -2*xi + 3*xi**2
    return (ui*M3 + vi*M5 + uj*M4 + vj*M6).flatten()


def displaced_profile(
        coord: Array,
        displ: Array,        #: Displacements
        vect : Array = None, #: Element orientation vector
        Q = None,
        npoints: int = 10,
        tangent: bool = False
    ):
    n = npoints
    #           (------ndm------)
    reps = 4 if len(coord[0])==3 else 2

    # 3x3 rotation into local system
    # Q = rotation(coord, vect)
    # Local displacements
    u_local = block_diag(*[Q]*reps)@displ
    # Element length
    L = np.linalg.norm(coord[-1] - coord[0])

    # longitudinal, transverse, vertical, section, elevation, plan
    li, ti, vi, si, ei, pi = u_local[:6]
    lj, tj, vj, sj, ej, pj = u_local[6:]

    Lnew  = L + lj - li
    xaxis = np.linspace(0.0, Lnew, n)

    plan_curve = elastic_curve(xaxis, [ti, pi, tj, pj], Lnew)
    elev_curve = elastic_curve(xaxis, [vi,-ei, vj,-ej], Lnew)

    local_curve = np.stack([xaxis + li, plan_curve, elev_curve])

    if tangent:
        plan_tang = elastic_tangent(xaxis, [ti, pi, tj, pj], Lnew)
        elev_tang = elastic_tangent(xaxis, [vi,-ei, vj,-ej], Lnew)

        local_tang = np.stack([np.linspace(0,0,n), plan_tang, elev_tang])
        return (
            Q.T@local_curve + coord[0][None,:].T,
            Q.T@local_tang
        )

    return Q.T@local_curve + coord[0][None,:].T


class FrameArtist:
    ndm: int
    ndf: int
    model: "FrameModel"
    canvas: "Canvas"

    def __init__(self, model_data, response=None, ndf=None, loc=None, vert=2, **kwds):
        config = Config()
        if "config" in kwds:
            apply_config(kwds.pop("config"), config)

        apply_config(kwds, config)
        self.config = config

        canvas_type = config.get("canvas", "matplotlib")
        if not isinstance(canvas_type, str):
            self.canvas = canvas_type
        elif canvas_type == "matplotlib":
            import sees.canvas.mpl
            self.canvas = sees.canvas.mpl.MatplotlibCanvas(**config["canvas_config"])
        elif canvas_type == "femgl":
            import sees.canvas.femgl
            self.canvas = sees.canvas.femgl.FemGlCanvas(self.model, **config["canvas_config"])
        elif canvas_type == "plotly":
            import sees.canvas.ply
            self.canvas = sees.canvas.ply.PlotlyCanvas(**config["canvas_config"])
        elif canvas_type == "gltf":
            import sees.canvas.gltf
            self.canvas = sees.canvas.gltf.GltfLibCanvas(**config["canvas_config"])
        elif canvas_type == "trimesh":
            import sees.canvas.tri
            self.canvas = sees.canvas.tri.TrimeshCanvas(**config["canvas_config"])
        else:
            raise ValueError("Unknown canvas type " + str(canvas_type))

        self.canvas.config = config


        self.ndm = 3

        if ndf is None:
            ndf = 6

        elif ndf == 3:
            self.ndm = 2

        if vert == 3:
            R = np.eye(3)
        else:
            R = np.array(((1,0, 0),
                          (0,0,-1),
                          (0,1, 0)))

        self._plot_rotation = R

        self.model = model = FrameModel(model_data, shift=loc, rot=R,
                                        **kwds.get("model_config", {}))


        # Create permutation matrix
        if model.ndf == 2 and model.ndm == 2:
            self.dofs2plot = R@np.array(((1, 0),
                                         (0, 1),
                                         (0, 0)))

        elif ndf == 3 and model.ndm == 2:
            self.dofs2plot = block_diag(*[R]*2)@np.array(((1,0, 0),
                                                          (0,1, 0),
                                                          (0,0, 0),

                                                          (0,0, 0),
                                                          (0,0, 0),
                                                          (0,0, 1)))

        else:
            self.dofs2plot = block_diag(*[R]*2)

        self.displ_states = {}

    def add_point_displacements(self, displ, scale=1.0, name=None):
        displ_array = self.displ_states[name]
        for i,n in enumerate(self.model["nodes"]):
            for dof in displ[n]:
                displ_array[i, dof] = 1.0

        displ_array[:,3:] *= scale/100
        displ_array[:,:3] *= scale
        return name

    def _add_displ_case(self, state, name=None, scale=1.0):

        if name in self.displ_states:
            self.displ_states[name].update(state)
        else:
            self.displ_states[name] = state

        return name


    def add_state(self, res_file, scale=1.0, only=None, type=None):

        if not isinstance(res_file, (dict, Array, State)):
            state = read_state(res_file, only=only,
                               model=self.model,
                               scale=scale,
                               transform=self.dofs2plot)
        else:
            state = res_file

        # If dict of dicts, assume its a collection of responses, 
        # otherwise, just a single response
        if isinstance(state, dict) and isinstance(next(iter(state.values())), dict):
            for k, v in state.items():
                self._add_displ_case(v, name=k, scale=scale)

        elif isinstance(state, dict):
            self._add_displ_case(BasicState(state, self.model,
                                            scale=scale,
                                            transform=self.dofs2plot))

        else:
            self._add_displ_case(state, scale=scale)


    def plot_origin(self, **kwds):
        xyz = np.zeros((3,3))
        uvw = self._plot_rotation.T*kwds.get("scale", 1.0)
        off = [[0, -kwds.get("scale", 1.0)/2, 0],
               [0]*3,
               [0]*3]

        self.canvas.plot_vectors(xyz, uvw, **kwds)

        for i,label in enumerate(kwds.get("label", [])):
            self.canvas.annotate(label, (xyz+uvw)[i]+off[i])

    def add_elem_data(self):

        N = 3
        for type in self.model.cell_prototypes():
            name = type["name"]
            ni = len(type["instances"])
            coords = np.zeros((ni*(N+1),self.ndm))
            coords.fill(np.nan)
            for i,crd in enumerate(self.model.cell_position(type["instances"])):
                coords[(N+1)*i:(N+1)*i+N,:] = np.linspace(*crd, N)

            # coords = coords.reshape(-1,4,N)[:,-N]
            coords = coords.reshape(-1,4,3)[:,-3]

            self.canvas.plot_hover(coords, data=type["properties"],
                                   style=NodeStyle(shape="sphere"), label=name)



    def plot_chords(self, state=None, layer=None):
        model = self.model
        ndm   = self.model["ndm"]

        N = 2
        do_lines = False
        lines = np.zeros((len(model["assembly"])*(N+1),3))
        lines.fill(np.nan)

        triangles = []
        nodes = model.node_position(state=state)

        for i,(tag,el) in enumerate(model["assembly"].items()):
            position = model.cell_position(tag, state)[[0,-1],:]
            if _is_frame(el):
                do_lines = True
                lines[(N+1)*i:(N+1)*i+N,:] = position #np.linspace(el["crd"][0], el["crd"][-1], N)

            else:
                triangles.extend(model.cell_triangles(tag))

        if do_lines:
            self.canvas.plot_lines(lines[:,:self.ndm])


        if len(triangles) > 0:
            self.canvas.plot_mesh(nodes, np.array(triangles))

    def draw_edges(self, state=None, style=None):
        model = self.model
        ndm   = self.model["ndm"]

        N = 2
        do_lines = False
        lines = np.zeros((len(model["assembly"])*(N+1),3))
        lines.fill(np.nan)

        indices = []
        nodes = model.node_position(state=state)

        for i,(tag,el) in enumerate(model["assembly"].items()):
            position = model.cell_position(tag, state)[[0,-1],:]
            if _is_frame(el):
                do_lines = True
                lines[(N+1)*i:(N+1)*i+N,:] = position

            else:
                idx = model.cell_exterior(tag)
                indices.append([*idx, idx[0]])

        if do_lines:
            self.canvas.plot_lines(lines[:,:self.ndm], style=style)

        if len(indices) > 0:
            self.canvas.plot_lines(nodes, indices=np.array(indices), style=style)

    def draw_filled(self, state=None):
        from sees.frame import extrude
        return extrude.draw_extrusions(self.model,
                                       canvas=self.canvas,
                                       state=state,
                                       options=self.config)

        sections = get_section_geometries(self.model, self.config)

        # N = 2
        N = 20 if displ is not None else 2

        I = 0
        coords = []
        triang = []
        for i,el in enumerate(self.model["assembly"].values()):
            # if int(el["name"]) < 30: continue
            try:
                sect = sections[el["name"]]
            except:
                if int(el["name"]) < 1e3:
                    sect = self.config["default_section"]
                else:
                    sect = np.array([[-48, -48],
                                     [ 48, -48],
                                     [ 48,  48],
                                     [-48,  48]])



    def plot_displaced_assembly(self, state=None, label=None):
        model = self.model
        N  = 10 if state is not None else 2
        ne = len(model["assembly"])
        coords = np.zeros((ne*(N+1), 3))
        coords.fill(np.nan)

        do_lines = False
        for i,(tag,el) in enumerate(model["assembly"].items()):
            # exclude zero-length elements
            if _is_frame(el) and state is not None:
                do_lines = True
                glob_displ = state.cell_array(tag).flatten()
                vect = None #np.array(el["trsfm"]["vecInLocXZPlane"])[axes]
                coords[(N+1)*i:(N+1)*i+N,:] = displaced_profile(model.cell_position(tag),
                                                                glob_displ,
                                                                Q=model.frame_orientation(tag),
                                                                npoints=N).T
            elif len(el["crd"]) == 2:
                do_lines = True
                coords[(N+1)*i:(N+1)*i+N,:] = np.linspace(*el["crd"], N)

        if do_lines:
            self.canvas.plot_lines(coords[:, :self.ndm], style=LineStyle(color="red"), label=label)

    def plot_nodes(self, state=None, data=None, label=None):
        coord = self.model.node_position(state=state)
        self.canvas.plot_nodes(coord[:,:self.ndm], label=label)

        if state is None:
            self.canvas.plot_hover(coord[:,:self.ndm],
                                   label="node",
                                   keys=["tag", "crd"],
                                   data=[[str(k), list(map(str, coord[i]))]
                                       for i,k in enumerate(self.model.iter_node_tags())])

    def add_triads(self):
        ne = len(self.model["assembly"])
        xyz, uvw = np.nan*np.zeros((2, ne, 3, 3))

        for i,el in enumerate(self.model["assembly"].values()):
            axes = self.model.frame_orientation(el["name"])
            if axes is None:
                continue

            scale = np.linalg.norm(el["crd"][-1] - el["crd"][0])/10
            coord = sum(i for i in el["crd"])/len(el["nodes"])
            xyz[i,:,:] = np.array([coord]*3)
            uvw[i,:,:] = scale*axes

        self.canvas.plot_vectors(xyz.reshape(ne*3,3), uvw.reshape(ne*3,3))

    def draw(self):
        if "reference" in self.config["show_objects"] \
            or len(self.displ_states) == 0:

            self.plot_chords()
#           self.draw_edges()
            try:
                self.add_elem_data()
            except Exception as e:
#               raise e
                print(e, file=sys.stderr)

        if "nodes" in self.config["show_objects"]:
            self.plot_nodes()

        if "origin" in self.config["show_objects"]:
            self.plot_origin(**self.config["objects"]["origin"])

        if "triads" in self.config["show_objects"]:
            self.add_triads()

        for layer, state in self.displ_states.items():
            self.plot_nodes(state=state, label=layer)

            if "chords" in self.config["show_objects"]:
#               self.draw_edges(state=state) #, style=state.line_style)
                self.plot_chords(state=state, layer=layer)

            self.plot_displaced_assembly(state=state, label=layer)

        if "extrude" in self.config["show_objects"]:
            displ = None
            if len(self.displ_states) == 1:
                displ = next(iter(self.displ_states.values()))

            try:
                self.draw_filled(displ)
            except Exception as e:
                # raise e
                print("Warning -- ", e, file=sys.stderr)

        self.canvas.build()
        return self

    def save(self, filename):
        self.canvas.write(filename)

    def repl(self):
        from opensees.repl.__main__ import OpenSeesREPL
        self.canvas.plt.ion()

        try:
            from IPython import get_ipython
            get_ipython().run_magic_line('matplotlib')
        except:
            pass

        repl = OpenSeesREPL()

        def plot(*args):
            if len(args) == 0:
                return self.draw()

            elif hasattr(self, "plot_"+args[0]):
                return getattr(self, "plot_"+args[0])(*args[1:])

            elif hasattr(self, args[0]):
                return getattr(self, args[0])(*args[1:])

        repl.interp._interp.createcommand("plot", plot)
        # repl.interp._interp.createcommand("show", lambda *args: self.canvas.show())
        repl.repl()




# Claudio Perez
import numpy as np
from collections import defaultdict

try:
    import orjson as json
except ImportError:
    import json

_EYE3 = np.eye(3)

_OUTLINES = {
    None:      None,
    "square":  np.array([[-1., -1.],
                         [ 1., -1.],
                         [ 1.,  1.],
                         [-1.,  1.]])/10,

    "tee":     np.array([[ 6.0,  0.0],
                         [ 6.0,  4.0],
                         [-6.0,  4.0],
                         [-6.0,  0.0],
                         [-2.0,  0.0],
                         [-2.0, -8.0],
                         [ 2.0, -8.0],
                         [ 2.0,  0.0]])/10
}



def read_model(filename:str, shift=None)->dict:

    if isinstance(filename, str) and filename.endswith(".tcl"):
        import opensees.tcl
        with open(filename, "r") as f:
            interp = opensees.tcl.exec(f.read(), silent=True, analysis=False)
        return interp.serialize()

    try:
        with open(filename,"r") as f:
            sam = json.loads(f.read())

    except TypeError:
        sam = json.loads(filename.read())

    return sam


class Model:
    def __iter__(self):
        # this method allows: nodes, cells = Model(mesh)
        return iter((self.nodes, self.cells))

    @property
    def nodes(self)->dict:
        pass

    @property
    def cells(self)->dict:
        pass

    def iter_nodes(self): ...

    def node_location(self, tag): ...

    def node_information(self, tag): ...

    def iter_cells(self, filt=None): ...

    def cell_type(self, tag):       ... # line triangle quadrilateral 

    def cell_exterior(self, tag):   ...

    def cell_interior(self, tag):   ...

    def cell_orientation(self, tag): ...

    def cell_position(self, tag, state):   ...

    def cell_outline(self,  tag):   ...


    def cell_information(self, tag): ...

class SolidModel:
    pass

class FrameModel:
    def __init__(self, sam:dict, shift = None, rot=None, **kwds):
        """
        Process OpenSees JSON output and return dict with the form:

            {<elem tag>: {"crd": [<coordinates>], ...}}
        """

        self._frame_outlines = None
        self._extrude_default = _OUTLINES[kwds.get("extrude_default", "square")]
        self._extrude_outline = _OUTLINES[kwds.get("extrude_outline", None)]

        ndm = 3
        R = np.eye(ndm) if rot is None else rot

        if shift is None:
            shift = np.zeros(ndm)
        else:
            shift = np.asarray(shift)

        #
        self._data = _from_opensees(sam, shift, R)# output

        self.ndm = self._data["ndm"]
        self.ndf = self._data["ndf"]

    def __getitem__(self, key):
        return self._data[key]

    def cell_nodes(self, tag=None):
        if tag is None:
            if not hasattr(self, "_cell_nodes"):
                self._cell_nodes = {k: e["nodes"] for k, e in self["assembly"].items()}
            return self._cell_nodes
        else:
            return self["assembly"][tag]["nodes"]

    def cell_indices(self, tag=None):
        if not hasattr(self, "_cell_indices"):
            self._cell_indices = {
                elem["name"]: tuple(self.node_indices(n) for n in elem["nodes"])
                for elem in self["assembly"].values()
            }

        if tag is not None:
            return self._cell_indices[tag]
        else:
            return self._cell_indices

    def cell_properties(self, tag=None):
        if tag is not None:
            return self["assembly"][tag]

    def cell_prototypes(self)->"iter":
        exclude_keys = {"type", "instances", "nodes",
                        "crd", "crdTransformation"}

        if not self["prototypes"]:
            elem_types = defaultdict(dict)

            for elem in self["assembly"].values():
                type = elem["type"]
                if type not in elem_types:
                    elem_types[type]["name"] = type
                    elem_types[type]["variants"] = []
                    elem_types[type]["instances"] = [elem["name"]]
                    elem_types[type]["properties"] = {
                        k: v for k,v in elem.items() if k not in exclude_keys
                    }
                else:
                    elem_types[type]["instances"].append(elem["name"])


            elem_types = list(elem_types.values())
        else:
            elem_types = [
                {
                    "name": f"{elem['type']}<{elem['name']}>",
                    "instances": [self["assembly"][i]["name"] for i in elem["instances"]],
                    "properties":  [
                        [str(v) for k,v in elem.items() if k not in exclude_keys]
                        #for _ in range(len(elem["instances"]))
                    ]*(len(elem["instances"])),
                    "coords": [self["assembly"][i]["crd"] for i in elem["instances"]],
                    "keys":   [k for k in elem.keys() if k not in exclude_keys]

                } for elem in self["prototypes"].get("elements", [])
            ]
        return elem_types

    def iter_node_tags(self):
        for tag in self["nodes"]:
            yield tag

    def iter_nodes(self, keys=None):
        pass

    def iter_cells(self, keys=None):
        pass

    def node_properties(self, tag=None)->dict:
        return self["nodes"][tag]

    def node_indices(self, tag=None):
        if not hasattr(self, "_node_indices"):
            self._node_indices = {
                tag: i for i, tag in enumerate(self["nodes"])
            }
        return self._node_indices[tag]

    def node_position(self, tag=None, state=None):

        if tag is None:
            pos = np.array([n["crd"] for n in self["nodes"].values()])
        else:
            pos = self["nodes"][tag]["crd"]

        if state is not None:
            pos += state.node_array(tag, dof=state.position)

        return pos

    def cell_position(self, tag, state=None):
        if isinstance(tag, list):
            return np.array([[
                    self.node_position(node, state) for node in self["assembly"][t]["nodes"]
                ] for t in tag
            ])
        else:
            return np.array([ self.node_position(node, state)
                              for node in self["assembly"][tag]["nodes"] ])

    def cell_exterior(self, tag):
        type = self["assembly"][tag]["type"].lower()

        if "frm" in type or "beamcol" in type:
            return self.cell_indices(tag)

        elif ("quad" in type or \
              "shell" in type and ("q" in type) or ("mitc" in type)):
            return self.cell_indices(tag)

        return []

    def cell_triangles(self, tag):
        type = self["assembly"][tag]["type"].lower()

        if "frm" in type or "beamcol" in type:
            return []

        elif ("quad" in type or \
              "shell" in type and ("q" in type) or ("mitc" in type)):
            nodes = self.cell_indices(tag)

            if len(nodes) == 4:
                return [[nodes[0], nodes[1], nodes[2]],
                        [nodes[2], nodes[3], nodes[0]]]
        return []

    def frame_orientation(self, tag):
        el  = self["assembly"][tag]
        xyz = el["crd"]

        v1  = xyz[-1] - xyz[0]
        L   = np.linalg.norm(v1)
        e1  = v1/L

        if "yvec" in el["trsfm"] and el["trsfm"]["yvec"] is not None:
            v2  = np.array(el["trsfm"]["yvec"])

        elif "vecInLocXZPlane" in el["trsfm"]:
            v13 =  np.atleast_1d(el["trsfm"]["vecInLocXZPlane"])
            v2  = -np.cross(e1,v13)

        else:
            return _EYE3

        e2 = v2 / np.linalg.norm(v2)
        v3 = np.cross(e1,e2)
        e3 = v3 / np.linalg.norm(v3)
        return np.stack([e1,e2,e3])

    def frame_outline(self, tag):
        if self._frame_outlines is None:
            self._frame_outlines = _get_frame_outlines(self)

        if self._extrude_outline is not None:
            return self._extrude_outline
        elif tag in self._frame_outlines:
            return self._frame_outlines[tag]
        else:
            return self._extrude_default


def _from_opensees(sam, shift, R):
    # Process OpenSees JSON format
    try:
        sam = sam["StructuralAnalysisModel"]
    except KeyError:
        pass

    geom = sam.get("geometry", sam.get("assembly"))

    try:
        #coord = np.array([R@n.pop("crd") for n in geom["nodes"]], dtype=float) + shift
        coord = np.array([R@n["crd"] for n in geom["nodes"]], dtype=float) + shift
        ndm = 3
    except:
        coord = np.array([R@[*n["crd"], 0.0] for n in geom["nodes"]], dtype=float) + shift
        ndm = 2

    nodes = {
        n["name"]: {**n, "crd": coord[i], "idx": i}
            for i,n in enumerate(geom["nodes"])
    }

#   ndm = len(next(iter(nodes.values()))["crd"])
    ndf = next(iter(nodes.values()))["ndf"]

    trsfm = {}
    for t in sam.get("properties", {}).get("crdTransformations", []):
        trsfm[int(t["name"])] = {
                k: val for k,val in t.items() if k != "vecInLocXZPlane"
        }
        trsfm[int(t["name"])]["vecInLocXZPlane"] = R@t["vecInLocXZPlane"]


    elems =  {
        e["name"]: dict(
        **e,
        crd=np.array([nodes[n]["crd"] for n in e["nodes"]], dtype=float),
        trsfm=trsfm[int(e["crdTransformation"])]
            if "crdTransformation" in e and int(e["crdTransformation"]) in trsfm
            else dict(yvec=R@e["yvec"] if "yvec" in e else None)
        ) for e in geom["elements"]
    }

    try:
        sections = {s["name"]: s for s in sam["properties"]["sections"]}
    except:
        sections = {}

    output = dict(nodes=nodes,
                  assembly=elems,
                  sam=sam,
                  sections=sections,
                  prototypes=sam.get("prototypes", {}),
                  ndm=ndm,
                  ndf=ndf
    )

    return output


def _add_section_shape(section, sections=None, outlines=None):
    from scipy.spatial import ConvexHull

    # Rotation to change coordinates from x-y to z-y
    R = np.array(((0,-1),
                  (1, 0))).T

    if "section" in section:
        # Treat aggregated sections
        if section["section"] not in outlines:
            outlines[section["name"]] = _add_section_shape(sections[section["section"]], sections, outlines)
        else:
            outlines[section["name"]] = outlines[section["section"]]

    elif "bounding_polygon" in section:
        outlines[section["name"]] = [R@s for s in section["bounding_polygon"]]

    elif "fibers" in section:
        #outlines[section["name"]] = _alpha_shape(np.array([f["coord"] for f in section["fibers"]]))
        points = np.array([f["coord"] for f in section["fibers"]])
        outlines[section["name"]] = points[ConvexHull(points).vertices]


def _get_frame_outlines(model):
    sections = {}
    for name,section in model["sections"].items():
        _add_section_shape(section, model["sections"], sections)

    outlines = {
        # TODO: For now, only using the first of the element's cross
        #       sections
        elem["name"]: np.array(sections[elem["sections"][0]])

        for elem in model["assembly"].values()
            if "sections" in elem and elem["sections"][0] in sections
    }

    return outlines

